#include "difftable.h"

/* update_q_diffs() - calculate the next wavefront of backward differences of 
 *                    fullstep values of q
 *  -2nd backward difference of q is calculated explicitly
 *  -all other backward differences are calculted with respect to this new
 *   value and all of the old values
 */
int update_q_diffs(Data *data)
{
  int i, j, k, natoms = data->natoms;
  double sum, u, dt = data->dt;
  double *m = data->mass, *f = data->f, *vel = data->vel;
  double **dif_q = data->diftab->dif_q;
  
  for(j=0; j<natoms; j++)
    {
      u = (dt*dt/m[j])*f[j]; // new 2nd backward difference of q (leapfrog)
      for(i=NUMSHADOWS; i>1; i--)
	{ for(sum = 0, k=2; k<i; k++) sum += dif_q[k][j];
	  dif_q[i][j] = u - sum; }
      dif_q[1][j] = dt*vel[j]; // new 1st backward difference of q (leapfrog)
    }
  return 0;
}


/* update_q_middiffs() - calculate the next wavefront of backward 
 *                       differences of midstep q values
 *  -1st backward difference of midstep q is calculated explicitly
 *  -all other backward differences are calculted with respect to this new
 *   value and all of the old values
 */
int update_q_middiffs(Data *data)
{
  int i, j, k, natoms = data->natoms;
  double sum, u, dt = data->dt;
  double *m = data->mass, *f = data->f, *vel = data->vel;
  double **dif_q = data->diftab->dif_q, **mdf_q = data->diftab->mdf_q;

  for(j=0; j<natoms; j++)
    {
      // new 1st backward difference of midstep values of q (leapfrog)
      u = dif_q[1][j] + (dt*dt/(2.0*m[j]))*f[j]; 
      for(i=NUMSHADOWS; i>0; i--)
	{ for(sum = 0, k=1; k<i; k++) sum += mdf_q[k][j];
	  mdf_q[i][j] = u - sum; }
    }

  return 0;
}


/* update_beta_diffs() - calculate the next wavefront of backward differences 
                         of fullstep values of beta
 *  -1st backward difference of beta is calculated explicitly
 *  -all other backward differences are calculted with respect to this new
 *   value and all of the old values
 */
int update_beta_diffs(Data *data, double beta)
{
  int i, k;
  double sum;
  double *dif_beta = data->diftab->dif_beta;

  for(i=NUMSHADOWS; i>0; i--)
    { for(sum = 0, k=1; k<i; k++) sum += dif_beta[k];
      dif_beta[i] = beta - sum; }

  return 0;
}


/* update_beta_middiffs() - calculate the next wavefront of backward 
 *                          differences of midstep beta values
 *  -1st backward difference of midstep beta is calculated explicitly
 *  -all other backward differences are calculted with respect to this new
 *   value and all of the old values
 */
int update_beta_middiffs(Data *data, double midbeta)
{
  int i, k;
  double sum;
  double *mdf_beta = data->diftab->mdf_beta;

  for(i=NUMSHADOWS; i>0; i--)
    { for(sum = 0, k=1; k<i; k++) sum += mdf_beta[k];
      mdf_beta[i] = midbeta - sum; }

  return 0;
}


/* calc_A_table() - calculate the values of the Aij used in the
 *                  formulas for interpolated shadow Hamiltonians
 *                  of order 2k with k even
 */
int calc_A_table(Data *data)
{
  int i, j, k, natoms = data->natoms;
  double dot, *m = data->mass, **A = data->diftab->A;
  double *dif_beta = data->diftab->dif_beta;
  double **dif_q = data->diftab->dif_q, **dif_p = data->diftab->dif_p;

  dif_q[0] = data->pos;
  dif_p[0] = data->vel;
  dif_beta[0] = 0; // shouldn't ever be used
  
  for(i=1; i<NUMDIFS; i++)
    {
      for(j=0; j<i; j++)
	{
	  // difference table stores velocity differences; hence
	  // mass must be multiplied in to get momentum.
	  for(dot=k=0; k<natoms; k++) dot += dif_q[i][k]*dif_p[j][k]*m[k];
	  A[i][j] = dot;
	  for(dot=k=0; k<natoms; k++) dot += dif_p[i][k]*dif_q[j][k]*m[k];
	  A[i][j] -= dot;
	}
      A[i][0] -= dif_beta[i];
    }

  for(i=1; i<NUMDIFS; i++)
    for(j=0; j<i; j++)
      A[i][j] /= 2.0*data->dt;

  return 0;
}


/* calc_B_table() - calculate the values of the Bij used in the
 *                  formulas for the interpolated shadow Hamiltonians
 *                  of order 2k with k odd
 */
int calc_B_table(Data *data)
{
  int i, j, k, natoms = data->natoms;
  double dot, *m = data->mass, **B = data->diftab->B;
  double **mdf_q = data->diftab->mdf_q;
  double **mdf_p = data->diftab->mdf_p;
  double *mdf_beta = data->diftab->mdf_beta;

  mdf_q[0] = data->midstep;
  mdf_p[0] = data->vel;
  mdf_beta[0] = 0; // shouldn't ever be used

  for(i=1; i<NUMDIFS; i++)
    {
      for(j=0; j<i; j++)
	{
	  // difference table stores velocity differences; hence
	  // mass must be multiplied in to get momentum.
	  for(dot=k=0; k<natoms; k++) dot += mdf_q[i][k]*mdf_p[j][k]*m[k];
	  B[i][j] = dot;
	  for(dot=k=0; k<natoms; k++) dot += mdf_p[i][k]*mdf_q[j][k]*m[k];
	  B[i][j] -= dot;
	}
      B[i][0] -= mdf_beta[i];
    }

  for(i=1; i<NUMDIFS; i++)
    for(j=0; j<i; j++)
      B[i][j] /= 2.0*data->dt;

  return 0;
}
