#include "force.h"
#include "truncated.h"

/*
 * doublewell.c contains code that is specific to a doublewell potential.
 * To use a new potential, you should only need to change things in this
 *  file, all other files should work fine as is.
 */

/*
 * Assign initial conditions here:
 *  Everything should scale to more than 1 atoms just fine.
 *  Energy & shadow energy will be computed correctly for any 
 *   number of atoms.
 *  Note that every atom in the system must be assigned an initial
 *   position, initial velocity, & a mass (ie. there are no defaults).
 */
int initial_conditions(Data *data)
{
  double *pos, *vel, *m;
  
  // number of atoms in the simulation
  data->natoms = 1; 

  // per atom parameters
  data->pos = (double *)malloc( data->natoms*sizeof(double) );
  data->midstep = (double *)malloc( data->natoms*sizeof(double) );
  data->vel = (double *)malloc( data->natoms*sizeof(double) );
  data->f = (double *)malloc( data->natoms*sizeof(double) );  
  data->mass = (double *)malloc( data->natoms*sizeof(double) );  

  pos = data->pos;
  vel = data->vel;
  m = data->mass;

  // initial conditions
  pos[0] = 0;
  vel[0] = 0.2;
  m[0] = 1;

  return 0;
}

/*
 * Forces are evaluated on a per atom basis.
 * force_eval is called separately for each atom at each iteration.
 */
int force_eval(Data *data, int i)
{
  int j, k, natoms = data->natoms;
  double *f = data->f, *pos = data->pos, u, sum;
  double **dif_p = data->diftab->dif_p, **mdf_p = data->diftab->mdf_p;

  u = f[i];  // (for interpolated shadow hamiltonian calculation)

  // DOUBLE WELL POTENTIAL F(q) = -Uq(q)  
  f[i] = -pos[i]*(pos[i]*pos[i] - 1);

  // BACKWARD DIFFERENCE TABLE UPDATE (FULL STEP P PORTION ONLY)
  u = (data->dt/(2.0*data->mass[i]))*(f[i] + u); // 1st difference of p
  for(j=NUMSHADOWS; j>0; j--)
    { for(sum=0, k=1; k<j; k++) sum += dif_p[k][i];
      dif_p[j][i] = u - sum; }
  
  // BACKWARD DIFFERENCE TABLE UPDATE (MIDSTEP P PORTION ONLY)
  u = (data->dt/data->mass[i])*f[i]; // 1st difference of p
  for(j=NUMSHADOWS; j>0; j--)
    { for(sum=0, k=1; k<j; k++) sum += mdf_p[k][i];
      mdf_p[j][i] = u - sum; }
 
  return 0;
}


/*
 * potential_eval is called once per iteration.
 */
int potential_eval(Data *data)
{
  int i, natoms = data->natoms;
  double *pos = data->pos, u;
  double sum = 0;
  
  // DOUBLE WELL POTENTIAL
  for(i=0; i<natoms; i++)
    { u = pos[i]*pos[i] - 1; sum += 0.25*u*u; }
  
  data->pe = sum;
  return 0;
}


/*
 * See comments in truncated.h
 *
 * Truncated hamiltonians depend on the particular potential being used.
 * These expressions are valid for a doublewell potential ONLY.
 *
 * The formulas for interpolated hamiltonians are *not* dependent on the
 *  potential, so those quantities will be valid whatever you change
 *  the potential function to.
 */

int truncated_eval(Data *data)     
{
  trunc_2_eval(data);
  trunc_4_eval(data);
  trunc_6_eval(data);
  trunc_8_eval(data);
  trunc_10_eval(data);

  return 0;
}

int trunc_2_eval(Data *data)
{
  double q = data->pos[0];
  double p = data->vel[0];
  double h = data->dt;
  data->trunc[S2] = 0.25 + 0.5*p*p - 0.5*q*q + 0.25*q*q*q*q;
  
  return 0;
}

int trunc_4_eval(Data *data)
{
  double q = data->pos[0];
  double p = data->vel[0];
  double h = data->dt;
  double p2 = p*p;
  double q2 = q*q;
  double q4 = q2*q2;
  double q6 = q4*q2;
  data->trunc[S4] = data->trunc[S2] + 
    h*h*( -p2/12.0 - q2/24.0 + p2*q2/4.0 + q4/12.0 - q6/24.0 );
  
  return 0;
}

int trunc_6_eval(Data *data)
{
  double q = data->pos[0];
  double p = data->vel[0];
  double h = data->dt;
  double p2 = p*p;
  double p4 = p2*p2;
  double q2 = q*q;
  double q4 = q2*q2;
  double q6 = q4*q2;
  double q8 = q4*q4;
  double h4 = h*h*h*h;
  data->trunc[S6] = data->trunc[S4] + 
    h4*( p2/60.0 - p4/120.0 + q2/240.0 - 3.0*p2*q2/20.0 - q4/48.0 
	  + p2*q4/5.0 + 7.0*q6/240.0 - q8/80.0 );
  
  return 0;
}

int trunc_8_eval(Data *data)
{
  double q = data->pos[0];
  double p = data->vel[0];
  double h = data->dt;
  double p2 = p*p;
  double p4 = p2*p2;
  double q2 = q*q;
  double q4 = q2*q2;
  double q6 = q4*q2;
  double q8 = q4*q4;
  double q10 = q8*q2;
  double h6 = h*h*h*h*h*h;
  data->trunc[S8] = data->trunc[S6] + 
    h6*( -p2/280.0 + p4/105.0 - q2/1680.0 + 43*p2*q2/560.0 - p4*q2/20.0
	  + 13*q4/1680.0 - 9*p2*q4/35.0 - 37.0*q6/1680.0 + 17*p2*q6/80.0
	  + 13.0*q8/560.0 - q10/120.0 );

  return 0;
}

int trunc_10_eval(Data *data)
{
  double q = data->pos[0];
  double p = data->vel[0];
  double h = data->dt;

  double p2 = p*p;
  double p4 = p2*p2;
  double p6 = p4*p2;

  double q2 = q*q;
  double q4 = q2*q2;
  double q6 = q4*q2;
  double q8 = q4*q4;
  double q10 = q8*q2;
  double q12 = q6*q6;

  double h2 = h*h;
  double h4 = h2*h2;
  double h8 = h4*h4;

  data->trunc[S10] = data->trunc[S8] + 
    h8*( p2/1260 - p4/140 + p6/420 + q2/10080 - 4*p2*q2/105 + 31*p4*q2/280
	 - 137*q4/40320 + 71*p2*q4/280 - 11*p4*q4/56 + 5*q6/288 - 69*p2*q6/140
	 - 15*q8/448 + 81*p2*q8/280 + 47*q10/1680 - 23*q12/2688);

  return 0;
}
