#include "force.h"

/*
 * henon.c contains code that is specific to the Henon-Heiles  potential.
 * To use a new potential, you should only need to change things in this
 *  file, all other files should work fine as is.
 */

/*
 * Assign initial conditions here:
 *  Everything should scale to more than 1 atoms just fine.
 *  Energy & shadow energy will be computed correctly for any 
 *   number of atoms.
 *  Note that every atom in the system must be assigned an initial
 *   position, initial velocity, & a mass (ie. there are no defaults).
 */
int initial_conditions(Data *data)
{
  double *pos, *vel, *m;
  
  // number of atoms in the simulation
  data->natoms = 2;

  // per atom parameters
  data->pos = (double *)malloc( data->natoms*sizeof(double) );
  data->midstep = (double *)malloc( data->natoms*sizeof(double) );
  data->vel = (double *)malloc( data->natoms*sizeof(double) );
  data->f = (double *)malloc( data->natoms*sizeof(double) );  
  data->mass = (double *)malloc( data->natoms*sizeof(double) );  

  pos = data->pos;
  vel = data->vel;
  m = data->mass;

  // initial conditions
  // x = 1/2 is chaotic; x = 1/3 is periodic
  pos[0] = 1.0/2.0; // x
  vel[0] = 0;       // u
  pos[1] = 0;       // y
  vel[1] = 0;       // v

  m[0] = 1;
  m[1] = 1;

  return 0;
}

/*
 * Forces are evaluated on a per atom basis.
 * force_eval is called separately for each atom at each iteration.
 */
int force_eval(Data *data, int i)
{
  int j, k, natoms = data->natoms;
  double *f = data->f, *pos = data->pos, u, x, y, sum;
  double **dif_p = data->diftab->dif_p, **mdf_p = data->diftab->mdf_p;

  x = data->pos[0]; 
  y = data->pos[1];
  
  u = f[i]; // store "old" value of the force
  if(i==0)
    f[i] = -(x + 2.0*x*y);   // -(d/dq1)U(q)
  else
    f[i] = -(y + x*x - y*y); // -(d/dq2)U(q)
  
  // BACKWARD DIFFERENCE TABLE UPDATE (FULLSTEP P PORTION ONLY)
  u = (data->dt/(2.0*data->mass[i]))*(f[i] + u); // 1st difference of p
  for(j=NUMSHADOWS; j>0; j--)
    { for(sum=0, k=1; k<j; k++) sum += dif_p[k][i];
      dif_p[j][i] = u - sum; }

  // BACKWARD DIFFERENCE TABLE UPDATE (MIDSTEP P PORTION ONLY)
  u = (data->dt/data->mass[i])*f[i]; // 1st difference of midstep of p
  for(j=NUMSHADOWS; j>0; j--)
    { for(sum=0, k=1; k<j; k++) sum += mdf_p[k][i];
      mdf_p[j][i] = u - sum; }
  
  return 0;
}


/*
 * potential_eval is called once per iteration.
 */
int potential_eval(Data *data)
{
  int i, natoms = data->natoms;
  double *pos = data->pos, u, x, y;

  x = pos[0];
  y = pos[1];

  // Henon-Heiles potential
  data->pe = 0.5*( x*x + y*y + 2.0*x*x*y - (2.0/3.0)*y*y*y );

  return 0;
}


/*
 * See comments in truncated.h
 *
 * Truncated shadow Hamiltonians are dependent on the Hamiltonian used.
 * These formulas are valid for the Henon-Heiles potential ONLY.
 *
 * The formulas for interpolated hamiltonians are *not* dependent on the
 *  potential, so those quantities will be valid whatever you change
 *  the potential function to.
 */
int truncated_eval(Data *data)     
{
  trunc_2_eval(data);
  trunc_4_eval(data);
  trunc_6_eval(data);
  trunc_8_eval(data);
  trunc_10_eval(data);
  trunc_12_eval(data);

  return 0;
}

int trunc_2_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double y3 = y2*y;

  data->trunc[S2] = (1.0/6.0) * ( 3*u2 + 3*v2 + 3*x2 + 6*x2*y 
				      + 3*y2 - 2*y3);
  
  return 0;
}

int trunc_4_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double y3 = y2*y;

  double x4 = x2*x2;
  double y4 = y2*y2;

  double h2 = data->dt * data->dt;

  data->trunc[S4] = data->trunc[S2] + 
    (1.0/24.0)*h2*( 2*u2 + 2*v2 + 8*u*v*x - x2 - x4 + 4*u2*y - 4*v2*y
		    - 6*x2*y - y2 - 2*x2*y2 + 2*y3 - y4 );
  
  return 0;
}

int trunc_6_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double y3 = y2*y;

  double x4 = x2*x2;
  double y4 = y2*y2;

  double y5 = y4*y;

  double h2 = data->dt * data->dt;
  double h4 = h2*h2;

  data->trunc[S6] = data->trunc[S4] + 
    (1.0/240.0)*h4*( -x2 - 5*x4 - 12*x2*y - 6*x4*y - y2 - 10*x2*y2 + 4*y3
		     - 4*x2*y3 - 5*y4 + 2*y5 + 8*u*v*x*(5 + 2*y) 
		     + 4*u2*(1 + 5*x2 + 5*y + 3*y2) 
		     + 4*v2*(1 + 3*x2 - 5*y + 5*y2) );
  
  return 0;
}

int trunc_8_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double x3 = x2*x;
  double y3 = y2*y;

  double x4 = x2*x2;
  double y4 = y2*y2;
  double u4 = u2*u2;
  double v4 = v2*v2;
  
  double y5 = y3*y2;
  
  double x6 = x3*x3;
  double y6 = y3*y3;

  double h2 = data->dt * data->dt;
  double h6 = h2*h2*h2;

  data->trunc[S8] = data->trunc[S6] + 
    (1.0/5040.0)*h6*( -12*u4 - 12*v4 - 3*x2 - 54*x4 - 7*x6 - 69*x2*y 
		      - 153*x4*y - 3*y2 - 108*x2*y2 - 111*x4*y2 + 23*y3
		      - 102*x2*y3 - 54*y4 + 39*x2*y4 + 51*y5 - 17*y6
		      + 336*u*v*x*(1 + x2 + y + y2) 
		      + 6*v2*(3 - 28*y + 66*y2 - 44*y3 + x2*(38 + 20*y))
		      + 6*u2*(3 - 4*v2 + 28*y + 38*y2
			      + 12*y3 + x2*(66 + 76*y)) );

  return 0;
}


int trunc_10_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double u3 = u2*u;

  double x3 = x2*x;
  double y3 = y2*y;

  double x4 = x2*x2;
  double y4 = y2*y2;
  double u4 = u2*u2;
  double v4 = v2*v2;
  
  double y5 = y3*y2;
  
  double x6 = x3*x3;
  double y6 = y3*y3;

  double y7 = y6*y;

  double h2 = data->dt * data->dt;
  double h4 = h2*h2;
  double h8 = h4*h4;

  data->trunc[S10] = data->trunc[S8] + 
    (1.0/10080.0)*h8*( -160*u3*v*x - x2 - 55*x4 
		       - 34*x6 - 42*x2*y - 276*x4*y - 60*x6*y - y2 
		       - 110*x2*y2 - 426*x4*y2 + 14*y3 - 184*x2*y3 
		       - 100*x4*y3 - 55*y4 + 114*x2*y4 + 92*y5 
		       - 20*x2*y5 - 70*y6 + 20*y7 
		       + 36*v4*(-1 + 2*y) - 4*u4*(9 + 14*y) 
		       - 8*u*v*x*(-31 + 12*v2 - 60*y - 126*y2 + 20*y3 
				  - 6*x2*(21 + 22*y)) 
		       + 4*v2*(2 + 45*x4 - 31*y + 132*y2 - 202*y3 
			       + 101*y4 - 6*x2*(-12 - 17*y + y2)) 
		       - 4*u2*(-2 - 63*x4 - 31*y - 72*y2 - 50*y3 
			       - 7*y4 + 6*v2*(3 + 2*y) - 
			       6*x2*(22 + 59*y + 37*y2)) );

  return 0;
}


int trunc_12_eval(Data *data)
{
  double x = data->pos[0];
  double u = data->vel[0];
  double y = data->pos[1];
  double v = data->vel[1];

  double x2 = x*x;
  double y2 = y*y;
  double u2 = u*u;
  double v2 = v*v;

  double u3 = u2*u;

  double x3 = x2*x;
  double y3 = y2*y;

  double x4 = x2*x2;
  double y4 = y2*y2;
  double u4 = u2*u2;
  double v4 = v2*v2;
  
  double y5 = y3*y2;
  
  double x6 = x3*x3;
  double y6 = y3*y3;

  double y7 = y6*y;

  double x8 = x4*x4;
  double y8 = y4*y4;

  double h2 = data->dt * data->dt;
  double h4 = h2*h2;
  double h10 = h4*h4*h2;

  data->trunc[S12] = data->trunc[S10] + 
    (1.0/221760.0)*h10*( -4*x2 - 615*x4 - 1120*x6 
			 - 133*x8 - 294*x2*y - 4902*x4*y - 4092*x6*y 
			 - 4*y2 - 1230*x2*y2 - 12504*x4*y2 - 2404*x6*y2 
			 + 98*y3 - 3268*x2*y3 - 6820*x4*y3 - 615*y4 
			 + 2736*x2*y4 - 1422*x4*y4 + 1634*y5 
			 - 1364*x2*y5 - 2136*y6 + 508*x2*y6 + 1364*y7 
			 - 341*y8 - 256*u3*v*x*(31 + 37*y) 
			 - 32*u4*(23 + 66*x2 + 82*y + 50*y2)
			 - 32*v4*(23 + 6*x2 - 110*y + 110*y2) 
			 + 16*u*v*x*(121 + 445*x4 + 411*y + 1390*y2 - 502*y3 
				     + 273*y4 + 16*v2*(-17 + 7*y) 
				     + 2*x2*(695 + 1719*y + 703*y2)) 
			 + 4*v2*(10 - 242*y + 1689*y2 - 4536*y3 + 5115*y4 
				 - 2046*y5 + x4*(2179 + 1546*y) 
				 + x2*(867 + 2488*y - 586*y2 + 876*y3)) 
			 + 4*u2*(10 + 242*y + 867*y2 + 1024*y3 + 209*y4 
				 + 78*y5 + x4*(3145 + 4358*y)) 
			 - 16*v2*(23 + 124*x2 + 42*y - 8*y2) 
			 + x2*(1689 + 8048*y + 11234*y2 + 3060*y3));

  return 0;
}


