#include "integrator.h"

int leapfrog(Data *data)
{
  int i, j, k, nsteps = data->nsteps, natoms = data->natoms;
  double *pos = data->pos, *vel = data->vel, *f = data->f, *m = data->mass;
  double *midstep = data->midstep;
  double beta, midbeta=0;
  double dot, dt = data->dt, half_dt = dt/2.0;

  for(k=0; k<nsteps; k++)
    {
      // HALF KICK : vel n+1/2
      for(j=0; j<natoms; j++) vel[j] += (half_dt/m[j])*f[j];

      // BACKWARD DIFFERENCE TABLE UPDATES (Q PORTIONS ONLY)
      update_q_middiffs(data); // uses F(n) & 1st diff. of q(n)
      update_q_diffs(data);    // uses F(n) & p(n+1/2)

      // Beta n+1/2
      for(dot=i=0; i<natoms; i++) dot += pos[i]*f[i];
      midbeta += beta = half_dt*(-dot - 2.0*data->pe);
      
      // BACKWARD DIFFERENCE TABLE UPDATE (MIDSTEP BETA PORTION ONLY)
      update_beta_middiffs(data, midbeta);

      // HALF-DRIFT : pos n+1/2
      for(j=0; j<natoms; j++) midstep[j] = pos[j] + half_dt*vel[j];

      // calculate Bij's
      calc_B_table(data); // 0th differences are midstep values

      // DRIFT : pos n+1
      for(j=0; j<natoms; j++) pos[j] += dt*vel[j];

      // FORCE : f n+1 & BACKWARD DIFFERENCE TABLE UPDATE (P PORTIONS ONLY)
      for(j=0; j<natoms; j++) force_eval(data,j); 
      
      // HALF KICK : vel n+1
      for(j=0; j<natoms; j++) vel[j] += (half_dt/m[j])*f[j];

      energy_eval(data); // potential energy n+1 (for use in Beta n+1)

      // Beta n+1
      for(dot=i=0; i<natoms; i++) dot += pos[i]*f[i];
      beta += midbeta = half_dt*(-dot - 2.0*data->pe);

      // BACKWARD DIFFERENCE TABLE UPDATE (FULLSTEP BETA PORTION ONLY)
      update_beta_diffs(data, beta);

      // backward difference table update completed (all portions)

      // calculate Aij's
      calc_A_table(data);

      // use Aij's & Bij's to calculate shadow hamiltonians
      shadow_hamiltonians_eval(data);
      truncated_eval(data);

      data->n++;
      if(k<NUMDIFS) show_some(data);
      else show_all(data);
    }

  return 0;
}


int energy_eval(Data *data)
{
  int i, natoms = data->natoms;
  double sum = 0;
  double *vel = data->vel, *m = data->mass;

  for(i=0; i<natoms; i++) sum += 0.5*m[i]*vel[i]*vel[i]; // kinetic energy

  data->energy = data->pe = 0; 
  potential_eval(data); // potential energy
  data->energy = sum + data->pe; // total = kinetic + potential

  return 0;
}


/* shadow_hamiltonians_eval() - calculate shadow energies
 *  -evaluates the interpolated shadow Hamiltonian formulas
 *   using the Aij & Bij terms (already computed) & the coefficients
 *   (initialized before start of leapfrog).
 */
int shadow_hamiltonians_eval(Data *data)
{
  int i, k, m, coefnum;
  double **A=data->diftab->A;

  // highest order differences are smallest
  // thus start summing with these, to minimize floating point error
  for(i=0; i<NUMSHADOWS; i++)
    {
      if(i%2==1) A=data->diftab->A;
      else A=data->diftab->B;
      data->sham[i] = 0;
      coefnum = -1;
      for(k=1; k<=i+1; k++)
	for(m=0; m<k; m++)
	  coefnum++;
      for(k=i+1; k>0; k--)
	for(m=k-1; m>=0; m--)
	  data->sham[i] += data->coef[i][coefnum--] * A[k][m];
    }
  
  return 0;
}
